import re
import datetime

dates = [datetime.date(2000, x+1, 1) for x in xrange(12)]
months = dict([(x.strftime("%b").upper(), (int(x.strftime("%m")))) for x in dates])

def getDate(year, monthName, day):
    day = int(day)
    month = int(months[monthName])
    return datetime.date(year, month, day)

def cleanMoney(amount):
    return amount.replace(' ', '').replace(',', '')

class Transaction(object):
    def __init__(self, date=None, desc='', amount='0.0'):
        self.date = date
        self.desc = desc
        self.amount = amount
    
class LineParser(object):
    def __init__(self, line, year, sectionType):
        self.line = line
        self.year = year
        if sectionType == 'deposit':
            self.regex = '(.*)(\s[0-9,]+\.[0-9]{2}$)'
        elif sectionType == 'withdrawal':
            self.regex = '(.*)(-\s[0-9,]+\.[0-9]{2}$)'
    def parse(self):
        words = self.line.split()
        date = None
        desc = ''
        amount = 0.0
        if words[0] in months.keys():
            date = getDate(self.year, words[0], words[1])
            desc = ''.join(self.line[2:])
            amount = 0.0
        if desc:
            line = desc
        else:
            line = self.line
        lastLine = re.match(self.regex,line)
        if lastLine:
            desc = lastLine.groups()[0]
            amount = cleanMoney(lastLine.groups()[1])
        if not desc:
            desc = self.line
        return (date, desc.strip(), amount)

class CheckSectionParser(object):
    def __init__(self, section, year):
        self.section = section
        self.year = year
        #self.section = ' '.join(section).split()
        # check section fields consist of check#, date, and amount
        self.fieldCount = 3
    def parseCheckLineTransactions(self, line):
        lineChecks = []
        regex = '([0-9]*)\*?\s?([%s)]{3})\s+([0-9]{2})\s([0-9,]+\.[0-9]{2})\s?' % '|'.join(months.keys())
        match = re.match("(%s)(%s)?" % (regex, regex), line)
        groups = match.groups()
        # first check
        date = getDate(self.year, groups[2], groups[3])
        desc = "Check # %s" % (groups[1] or 'unknown')
        transaction = Transaction(date, desc, cleanMoney('-'+groups[4]))
        lineChecks.append(transaction)
        # see if there is a second one on this line
        if groups[5]:
            date = getDate(self.year, groups[7], groups[8])
            desc = "Check # %s" % (groups[6] or 'unknown')
            transaction = Transaction(date, desc, cleanMoney('-'+groups[9]))
            lineChecks.append(transaction)
        return lineChecks
    def getTransactions(self):
        transactions = []
        for line in self.section:
            words = line.split()
            if words[0] in months or words[1] in months:
                transactions.extend(self.parseCheckLineTransactions(line))
        #import pdb;pdb.set_trace()
        return transactions

class SectionParser(object):
    def __init__(self, section, year, sectionType):
        self.section = section
        self.sectionType = sectionType
        self.year = year
    def getTransaction(self):
        transaction = None
        isSameTransaction = True
        while isSameTransaction:
            try:
                line = self.section.pop(0)
            except IndexError:
                break
            if line:
                lp = LineParser(line, self.year, self.sectionType)
            date, desc, amount = lp.parse()
            if date:
                transaction = Transaction()
                transaction.date = date
            if transaction:
                transaction.desc += desc
            if amount:
                transaction.amount = amount
                isSameTransaction = False
        return transaction
    def getTransactions(self):
        transactions = []
        while self.section:
            transaction = self.getTransaction()
            if transaction:
                transactions.append(transaction)
        return transactions

class FileParser(object):
    def __init__(self, filename, year):
        self.filename = filename
        fh = open(filename)
        self.data = fh.read().replace('\xe2\x88\x92', '-')
        self.lines = self.data.split('\r')
        self.year = year
        self.deposits = None
        self.withdrawals = None
    def find(self, text):
        count = 0
        for line in self.lines:
            if text in line:
                return (count, line.find(text))
            count += 1
        return (None, None)
    def getDepositSection(self):
        depositStart = 'DEPOSITS AND CREDITS'
        depositEnds = ['WITHDRAWALS AND DEBITS']
        startIndex, column = self.find(depositStart)
        indexChoices = []
        for possibleEnd in depositEnds:
            row, column = self.find(possibleEnd)
            indexChoices.append(row)
        indexChoices.sort()
        endIndex = indexChoices[0]
        return self.lines[startIndex:endIndex]
    def getWithdrawalSection(self):
        withdrawalStart = 'WITHDRAWALS AND DEBITS'
        withdrawalEnds = ['CHECKS PAID', 'DAILY BALANCE SUMMARY']
        startIndex, column = self.find(withdrawalStart)
        if not startIndex:
            raise
        indexChoices = []
        for possibleEnd in withdrawalEnds:
            row, column = self.find(possibleEnd)
            if row:
                indexChoices.append(row)
        indexChoices.sort()
        endIndex = indexChoices[0]
        return self.lines[startIndex:endIndex]
    def getChecksPaidSection(self):
        checksStart = 'CHECKS PAID'
        checksEnds = ['DAILY BALANCE SUMMARY']
        startIndex, column = self.find(checksStart)
        if not startIndex:
            return []
        indexChoices = []
        for possibleEnd in checksEnds:
            row, column = self.find(possibleEnd)
            indexChoices.append(row)
        indexChoices.sort()
        endIndex = indexChoices[0]
        return self.lines[startIndex:endIndex]
    def parseDeposits(self):
        data = self.getDepositSection()
        sp = SectionParser(data, self.year, 'deposit')
        self.deposits = sp.getTransactions()
    def parseWithdrawals(self):
        data = self.getWithdrawalSection()
        sp = SectionParser(data, self.year, 'withdrawal')
        self.withdrawals = sp.getTransactions()
        data = self.getChecksPaidSection()
        cp = CheckSectionParser(data, self.year)
        checks = cp.getTransactions()
        #import pdb;pdb.set_trace()
        self.withdrawals.extend(checks)
    def parse(self):
        self.parseDeposits()
        self.parseWithdrawals()
    def getDeposits(self):
        if not self.deposits:
            self.parseDeposits()
            self.deposits.sort(lambda a,b: cmp(a.date, b.date))
        return self.deposits
    def getWithdrawals(self):
        if not self.withdrawals:
            self.parseWithdrawals()
            self.withdrawals.sort(lambda a,b: cmp(a.date, b.date))
        return self.withdrawals
    def getTotalDeposits(self):
        total = 0.0
        for dep in self.getDeposits():
            total += float(dep.amount)
        return total
    def getTotalWithdrawals(self):
        total = 0.0               
        for withd in self.getWithdrawals():
            total += float(withd.amount)
        return total

#filename = '/Volumes/adytum/Finances/BankStatements/2005/03.16.txt'
#filename = '/Volumes/adytum/Finances/BankStatements/2005/04.18.txt'
#filename = '/Volumes/adytum/Finances/BankStatements/2005/06.16.txt'
#filename = '/Volumes/adytum/Finances/BankStatements/2005/07.19.txt'
#filename = '/Volumes/adytum/Finances/BankStatements/2005/08.16.txt'
#filename = '/Volumes/adytum/Finances/BankStatements/2005/09.19.txt'
#filename = '/Volumes/adytum/Finances/BankStatements/2005/10.19.txt'
#filename = '/Volumes/adytum/Finances/BankStatements/2005/11.17.txt'
filename = '/Volumes/adytum/Finances/BankStatements/2005/12.16.txt'
p = FileParser(filename, 2005)
print "Deposits"
print "========"
for dep in p.getDeposits():
    print [dep.date, dep.desc, dep.amount]
print "Withdrawals"
print "==========="
for withd in p.getWithdrawals():
    print [withd.date, withd.desc, withd.amount]
print "Total deposits: %s" % p.getTotalDeposits()
print "Total deposits: %s" % p.getTotalWithdrawals()
