import re
import os


if os.name == 'posix':
    TRACE_BIN = '/usr/sbin/traceroute'
else:
    TRACE_BIN = None

class Trace(object):
    """
    >>> from traceroute import Trace
    >>> t = Trace(host='www.adytum.us')
    >>> t.call()
    >>> len(t.results.getHosts()) > 0
    """
    def __init__(self, host='127.0.0.1', useDNS=False, queries='1'):
        self.host = host
        self.numeric = not useDNS
        self.queries = queries
        self._setParams()
        self._setCommand()

    def _setParams(self):
        self.params = ''
        if self.numeric:
            self.params +=' -n'
            self.mode = 'numeric'
        else:
            self.mode = 'alpha'
        if self.queries:
            self.params += ' -q %s' % self.queries

    def _setCommand(self):
        self.cmd = '%s %s %s' % (TRACE_BIN, self.params, self.host)

    def _executeCommand(self):
        p = os.popen(self.cmd)
        data = p.readlines()
        err = p.close()
        if err:
            raise RuntimeError, '%s failed w/ exit code %d' % (cmd, err)
        self.results = ParseOutput(data, self.mode)

    run = _executeCommand


class ParseError(Exception):
    """
    There was a problem parsing the data.
    """


class ParseLine(object):
    """
    For a single-query traceroute, a hop can have either:
        <hop index> <host> <latency>
    or
        <hop index> <asterisk>

    For a two-query traceroute, a hop can have one of the following:
        <hop index> <host> <latency> <latency>
        <hop index> <host1> <latency> <host2> <latency>
        <hop index> <host> <asterisk> <latency>
        <hop index> <host> <latency> <asterisk>
        <hop index> <asterisk> <asterisk>

    In general, a line consists of a set of one or more host responses, where a
    host response can be of the following forms:
        (<host>, <latency>)
        (<latency>)
        (<asterisk>)

    A single-element, latency-only host response can only come after a
    double-element host response. A single-element asterisk-only host response
    can either come after a double-element host response, or it can be on a hop
    that has no host defined.

    A traceroute hop line with one query will have at most three parts and at
    least two parts. A hop line with two queries will have at most five parts
    and at least three parts. A hop line with three queries will have at most
    seven parts and at least four parts. In general, a line will have the
    following parts:
        max: q * 2 + 1
        min: q + 1
    where "q" is the number of queries. The maximum part counts occur when each
    query gets a different host; the minimum part counts occur when no host is
    returned for any query.
    """
    domainPattern = re.compile(
        "([a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?\.)*"
        "[a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?")
    ipRegex = (
        "(?P<ipAddress>"
        "([01]?\d\d?|2[0-4]\d|25[0-5])\."
        "([01]?\d\d?|2[0-4]\d|25[0-5])\."
        "([01]?\d\d?|2[0-4]\d|25[0-5])\."
        "([01]?\d\d?|2[0-4]\d|25[0-5]))")
    ipPattern = re.compile(ipRegex)
    ipMapPattern = re.compile("\(%s\)" % ipRegex)
    latencyPattern = re.compile("[0-9]+\.[0-9]+")

    def __init__(self, line):
        self.line = line
        self.parsed = []
        self.parse()

    def __getitem__(self, index):
        return self.parsed[index]

    def __iter__(self):
        return iter(self.parsed)

    def __len__(self):
        return len(self.parsed)

    def __repr__(self):
        return str(self.parsed)

    def _findData(self, key):
        # We want to start searching with the most recent, and work our way
        # backwards.
        for data in reversed(self):
            value = data.get(key)
            if value:
                return value
        return ""

    def _consume(self, hopIndex, parts):
        domainName = ""
        ipAddress = ""
        latency = ""
        data = {
            "hopIndex": hopIndex,
            "domainName": domainName,
            "ipAddress": ipAddress,
            "latency": latency}
        if parts[0] == "*":
            parts.pop(0)
            return data, parts
        # Let's look ahead one and see if we've got a domain/ip pair.
        match = self.ipMapPattern.match(parts[1])
        if match:
            data["domainName"] = parts.pop(0)
            data["ipAddress"] = match.group("ipAddress")
            parts.pop(0)
        elif self.ipPattern.match(parts[0]):
            data["ipAddress"] = parts.pop(0)
        if self.latencyPattern.match(parts[0]):
            data["latency"] = parts.pop(0)
            # Then dump the "ms" units.
            parts.pop(0)
        # If we don't have a domain name or ip, let's check to see if a former
        # query obtained one.
        if not data.get("domainName"):
            data["domainName"] = self._findData("domainName")
        if not data.get("ipAddress"):
            data["ipAddress"] = self._findData("ipAddress")
        return data, parts

    def parse(self):
        last_host_seen = ""
        parts = re.split("\s+", self.line.strip())
        maxCount = len(parts)
        hopIndex = parts.pop(0)
        count = 0
        while parts:
            try:
                data, parts = self._consume(hopIndex, parts)
            except IndexError:
                raise ParseError("Invalid traceroute data: incomplete line.")
            self.parsed.append(data)
            count += 1
            if count > maxCount:
                raise ParseError("Invalid traceroute data; could not parse.")

    def getHopIndex(self):
        return int(self[0].get("hopIndex"))

    def getIPs(self):
        return [data.get("ipAddress") for data in self
                if data.get("ipAddress")]

    def getDomains(self, subdomains=0):
        return [data.get("domainName") for data in self 
                if data.get("domainName")]

    def getLatencies(self):
        return [float(data.get("latency")) for data in self
                if data.get("latency")]

    def getAverageLatency(self):
        latencies = self.getLatencies()
        return sum(latencies) / len(latencies)


class ParseOutput(object):

    def __init__(self, dataList, mode):
        # Get rid of the summary line.
        self.data = dataList[1:]
        self.mode = mode

    def getHopCount(self):
        return len(self.data)

    def getNextHop(self, thisIndex, hosts):
        nextIndex = thisIndex + 1
        try:
            return hosts[nextIndex]
        except IndexError:
            return

    def getIPs(self):
        ipGroups = []
        for line in self.data:
            ips = ParseLine(line).getIPs()
            if ips:
                ipGroups.append(ips)
        return ipGroups

    def getLinkedIPs(self):
        """
        This method provides a tuple of tuples of the pattern
            ((1,2), (2,3), (3,4) ... (n-1, n))
        for the purpose of using in graphs a la pydot.
        """
        nodes = []
        hosts = self.getIPs()
        for groupIndex, ipGroup in enumerate(self.getIPs()):
            nextHop = self.getNextHop(groupIndex, hosts)
            if not nextHop:
                continue
            for ip in ipGroup:
                for nextIp in nextHop:
                    node = (ip, nextIp)
                    try:
                        alreadyThere = nodes.index(node)
                    except ValueError:
                        nodes.append(node)
        return nodes

    def getLinkedDomains(self, limitSubdomains=0):
        """
        This does a similar thing as the getLinkedHosts() method. In fact,
        if the .mode attribute is numeric, it just calls getLinkedHosts().
        However, if not, pull out host and subdomain from the FQDN and only
        retain domain-level name info.
        """
        if self.mode == 'numeric':
            return self.getLinkedHosts()
        else:
            uniquer = {}
            hosts = self.getHosts()
            for host in hosts:
                domain = '.'.join(host.split('.')[-2:])
                uniquer.update({domain:hosts.index(host)})
            hosts = zip(uniquer.values(), uniquer.keys())
            hosts.sort()
            return [(y, hosts[hosts.index((x,y))+1][1]) for x,y in hosts
                    if len(hosts) > hosts.index((x,y)) + 1]
