from genshi.output import DocType
from genshi.template import TemplateLoader

from trac.web.api import *
from trac.config import default_dir
from trac.web.main import get_environments, _open_environment
from trac.util.text import to_unicode

def getNameAndCategory(hidden_name):
    if hidden_name.startswith('zi__'):
        cat = 'Internal'
        name = hidden_name.replace('zi__', '')
    elif hidden_name.startswith('zp__'):
        cat = 'Customers/Partners'
        name = hidden_name.replace('zp__', '')
    elif hidden_name.startswith('zd__'):
        cat = 'Deprecated'
        name = hidden_name.replace('zd__', '')
    elif hidden_name.startswith('zz__'):
        cat = 'Miscellany'
        name = hidden_name.replace('zz__', '')
    elif hidden_name.startswith('zc__'):
        cat = 'Community'
        name = hidden_name.replace('zc__', '')
    else:
        cat = 'Open Source'
        name = hidden_name
    return {'name':name, 'category':cat}

def sortProjectData(x, y):
    result = cmp(x['hiddenName'].lower(), y['hiddenName'].lower())
    if result == 0:
        result = cmp(x['name'].lower(), y['name'].lower())
    return result

def makeGroups(projects):
    groups = []
    for project in projects:
        name = project.get('category')
        data = {
            'name': name,
            'projects': [],
        }
        if name not in [ x.get('name') for x in groups ]:
            groups.append(data)
    for group in groups:
        for project in projects:
            if group.get('name') == project.get('category'):
                group['projects'].append(project)
    return groups

class ProjectsIndexOneX(object):
    """
    This class can be used to override the default trac project listing for
    trac 1.x. To do so, you will need to re-install the trac python libs after
    making the following change to trac/web/main.py:

        * Rename main.send_project_index() to something else
        * Create a new main.send_project_index(), containing the following
          code:

        def send_project_index(environ, start_response, parent_dir=None,
                               env_paths=None):
            from pkg_resources import require
            require('PyAdytum >= 1.0.6')
            from adytum.plugins.trac.multiprojects import ProjectsIndexOneX as ProjectsIndex
            index = ProjectsIndex(environ, start_response, parent_dir, env_paths)
            index.render() 

    As for the rendering of the data presented here, the following is an
    example using the Trac Genshi templating engine:

        <!DOCTYPE html
            PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
            "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml"
              xmlns:py="http://genshi.edgewall.org/"
              xmlns:xi="http://www.w3.org/2001/XInclude">
          <head>
            <title>Adytum Projects</title>
          </head>

          <body>
            <h1>Adytum Projects</h1>
            <div py:for="group in groups" py:choose="">
                <h2>$group.name</h2>
                <ul>
                  <li py:for="project in group.projects" py:choose="">
                    <a py:when="project.href" href="$project.href"
                       title="$project.description">$project.name</a>
                    - $project.description
                  </li>
                </ul>
            </div>
          </body>
        </html>


    """

    def __init__(self, environ, start_response, parent_dir=None,
        env_paths=None):
        self.environ = environ
        self.parentDir = parent_dir
        self.envPaths = env_paths
        self.req = Request(self.environ, start_response)

    def getProjects(self):
        href = Href(self.req.base_path)
        projects = []
        for env_name, env_path in get_environments(self.environ).items():
            try:
                env = _open_environment(env_path,
                                        run_once=self.environ['wsgi.run_once'])
                nameCat = getNameAndCategory(env.project_name)
                proj = {
                    'hiddenName': env.project_name,
                    'name': nameCat.get('name'),
                    'category': nameCat.get('category'),
                    'description': env.project_description,
                    'href': href(env_name)
                }
                projects.append(proj)
            except Exception, e:
                #proj = {'name': env_name, 'description': to_unicode(e)}
                pass
        projects.sort(sortProjectData)
        return makeGroups(projects)

    def getProjectsListing(self):
        pass

    def render(self):
        loadpaths = [default_dir('templates')]
        if self.req.environ.get('trac.env_index_template'):
            tmpl_path, template = os.path.split(self.req.environ['trac.env_index_template'])
            loadpaths.insert(0, tmpl_path)
        else:
            template = 'index.html'

        data = {}
        if self.req.environ.get('trac.template_vars'):
            for pair in self.req.environ['trac.template_vars'].split(','):
                key, val = pair.split('=')
                self.req.hdf[key] = val
                data[key] = val
                
        if self.parentDir and not self.envPaths:
            self.envPaths = dict([(filename, os.path.join(self.parentDir, filename))
                              for filename in os.listdir(self.parentDir)])
        try:
            data['groups'] = self.getProjects()
            markuptemplate = TemplateLoader(loadpaths).load(template)
            stream = markuptemplate.generate(**data)
            output = stream.render('xhtml', doctype=DocType.XHTML_STRICT)
            self.req.send(output, 'text/html')
        except RequestDone:
            pass

class ProjectsIndex(object):
    '''
    This class should be used inside the trac mod_python 
    handler 'ModPythonHandler.py', specifically, in the
    'send_project_index()' function.

    For example, this file:
        /usr/local/lib/python2.3/site-packages/trac/ModPythonHandler.py

    It should be used in the following manner:

    def send_project_index(req, mpr, dir):
        from adytum.plugins.trac import MultiProjectIndex

        text = MultiProjectIndex()
        text.setTitle('Available Projects')
        text.setBaseURL(mpr.idx_location)
        text.setDirs(dir)

        req.content_type = text.getContentType()
        req.write(text.render())

    '''
    def __init__(self, content_type='text/html'):
        self.content_type = content_type
        self.location = ''
        self.dirs = ''
        self.title = ''
        self.header_file = ''
        self.footer_file = ''
        self.base_url = ''
        
    def setTitle(self, title):
        self.title = title

    def setLocation(self, location):
        self.location = location

    def setBaseURL(self, url):
        self.base_url = url

    def setDirs(self, dir):
        import os
        dirs = [ (x.lower(), x) for x in os.listdir(dir) ]
        dirs.sort()
        self.dirs = [ x[1] for x in dirs ]
        self.base_dir = dir
        default_header = 'header.tmpl'
        default_footer = 'footer.tmpl'
        self.header_file = '%s/%s' % (dir, default_header)
        self.footer_file = '%s/%s' % (dir, default_footer)        

    def getContentType(self):
        return self.content_type

    def getHeader(self):
        '''
        '''
        text = file(self.header_file, 'r').read()
        return text.replace('%%TITLE%%', self.title)

    def getProjectListing(self):
        '''
        We do some strange, custom stuff here in order to get the display
        the way we want it.

            * Internal projects are marked with zi__ProjectName
            * Customer/Partner projects are marked with zp__ProjectName
        '''
        import os
        from ConfigParser import ConfigParser

        cfg = ConfigParser()
        cat_heading = '<h3>%s</h3>'                 # takes the project grouping as a parameter
        heading = '<ul>'
        link = '<li><a href="%s">%s</a> - %s</li>'  # takes the mpr.idx_location/location as the URL
                                                    # and project as the display link, then the desc
        footing = '</ul>'

        listing_data = []
        for project in self.dirs:
            proj_dir = os.path.join(self.base_dir, project)
            #output += '(debug) proj_dir: %s<br/>' % proj_dir
            proj_conf = os.path.join(proj_dir, 'conf/trac.ini')
            #output += '(debug) proj_conf: %s<br/>' % proj_conf
            if os.path.isfile(proj_conf):
                cfg.read(proj_conf)
                url = '%s/%s' % (self.base_url, project)
                hidden_name = cfg.get('project', 'name')
                nameCat = getNameAndCategory(hidden_name)
                name = nameCat.get('name')
                cat = nameCat.get('category')
                descr = cfg.get('project', 'descr')
                listing_data.append((hidden_name, cat, name, url, descr))
        listing_data.sort()
        output = heading
        cat_last = None
        for hidden_name, cat, name, url, descr in listing_data:
            if cat != cat_last:
                output += cat_heading % cat
            output += link % (url, name, descr)
            cat_last = cat
        output += footing
        return output

    def getFooter(self):
        '''
        '''
        return file(self.footer_file, 'r').read()

    def render(self):
        '''
        '''
        output = self.getHeader()
        output += self.getProjectListing()
        output += self.getFooter()
        return output

def getCGIHTML():
    index = MultiProjectIndex()
    index.setTitle('Adytum Projects')
    index.setBaseURL('http://projects.adytum.us/tracs')
    index.setDirs('/srv/www/projects')
    return index.render()
