import os
import re
import time
import calendar
try:
    set()
except ValueError:
    from sets import Set as set
from datetime import datetime

from storm.locals import *

from matplotlib import axes
from matplotlib.figure import Figure
from matplotlib.ticker import Locator, Formatter
from matplotlib.collections import LineCollection
from matplotlib.backends.backend_agg import FigureCanvasAgg

second = 1
minute = 60
hour = 60 * minute
day = 24 * hour
week = 7 * day
month = 30.4375 * day
year = 365.25 * day

def convertDT2Epoch(dtObj):
    return time.mktime(dtObj.timetuple())

def convertEpoch2DT(epoch):
    struct = time.localtime(epoch)
    return datetime(*struct[:6])

def formatEpoch(epoch):
    return convertEpoch2DT(epoch).strftime('%Y.%m.%d')

def dateFromString(dateStr):
    parts = [int(x) for x in dateStr.split('-')]
    return datetime(*parts)

def makeEpochDateRange(dateRange):
    """
    dateRange can be:
    
        1) a string, either representing a single date or a range of
        dates, e.g., '2007-12-31' or '2007-12-20, 2007-12-31', or
        2) a tuple/list of epoch dates.
        
    returns a two-element list of epoch times representing the date range.
    """
    if isinstance(dateRange, list):
        return dateRange
    dates = [dateFromString(x.strip()) for x in dateRange.split(',')]
    epochDates = [convertDT2Epoch(x) for x in dates]
    if len(epochDates) == 1:
        epochDates.append(epochDates[0] + day - second)
    elif len(epochDates) == 2:
        epochDates[1] = epochDates[1] + day - second
    return epochDates

def getLastTimePeriod(epochRange):
    """
    epochRange is two dates representing a time period span in epoch time

    Given an epoch range, the time covered in that range is calcualted and an
    epoch range is returned that spans the period of time prior to the given 
    range.
    
    This function is useful when:
        1) one is considering a sequece of ranges, and
        2) the ranges cover equal time spans.
    """
    delta = epochRange[1] - epochRange[0]
    return [epochRange[0] - delta, epochRange[0]]

def getBranches(changedFiles):
    branches = []
    for change in changedFiles:
        parts = change.split('/')
        if change.startswith('branches'):
            try:
                branches.append(parts[1])
            except IndexError:
                # no branch name, skip it
                pass
        elif change.startswith('trunk'):
            branches.append(parts[0])
    return branches

def parseTicketsFromBranches(branches):
    tickets = []
    for branch in branches:
        parts = branch.split('-')
        numbers = parts[-2:]
        if u'trunk' in numbers:
            continue
        if numbers[0].strip('abcdefghijklmnopqrstuvwxyz'):
            ticket = numbers[0]
        else:
            try:
                ticket = numbers[1]
            except IndexError:
                continue
        tickets.append('#'+ticket)
    return tickets

def titleFromCamelCase(camelCase):
    camelIndices =  [ i for i,x in enumerate(camelCase) if x.isupper()]
    for bump, index in enumerate(camelIndices):
        actualIndex = index + bump
        camelCase = camelCase[:actualIndex] + ' ' + camelCase[actualIndex:]
    return camelCase.title()

class Revision(object):
    """
    A storm ORM object representing a row in the trac 'revsion' SQLite table.
    """
    __storm_table__ = 'revision'
    rev = Unicode(primary=True)
    time = Int()
    author = Unicode()
    message = Unicode()

    def getDateTime(self):
        try:
            self.dt
        except AttributeError:
            self.dt = convertEpoch2DT(self.time)
        return self.dt

class NodeChange(object):
    """
    A storm ORM object representing a row in the trac 'node_change' SQLite
    table.
    """
    __storm_table__ = 'node_change'
    __storm_primary__ = ('rev', 'path', 'change_type')
    rev = Unicode()
    path = Unicode()
    node_type = Unicode()
    change_type = Unicode()
    base_path = Unicode()
    base_rev = Unicode()

Revision.nodeChanges = ReferenceSet(Revision.rev, NodeChange.rev)

class TicketChange(object):
    """
    A storm ORM object representing a row in the trac 'ticket_change' SQLite
    table.
    """
    __storm_table__ = 'ticket_change'
    __storm_primary__ = ('ticket', 'time', 'field')
    ticket = Unicode()
    time = Int()
    author = Unicode()
    field = Unicode()
    oldvalue = Unicode()
    newvalue = Unicode()

class TracStore(Store):
    """
    dbName = '/Volumes/duncan/Work/Divmod/Exponent/trac.db'
    s = TracStore(dbName)
    """
    def __init__(self, dbName):
        db = create_database('sqlite:%s' % dbName)
        super(TracStore, self).__init__(db)

    def getRevisions(self, dateRange, userName=''):
        epochRange = makeEpochDateRange(dateRange)
        if not userName:
            return self.find(
                Revision,
                Revision.time >= epochRange[0],
                Revision.time <= epochRange[1])
        else:
            return self.find(
                Revision,
                Revision.time >= epochRange[0],
                Revision.time <= epochRange[1],
                Revision.author == userName)

    def getTickets(self, dateRange, userName=''):
        epochRange = makeEpochDateRange(dateRange)
        if not userName:
            return self.find(
                TicketChange,
                TicketChange.time >= epochRange[0],
                TicketChange.time <= epochRange[1])
        else:
            return self.find(
                TicketChange,
                TicketChange.time >= epochRange[0],
                TicketChange.time <= epochRange[1],
                TicketChange.author == userName)

class TracStats(object):
    """
    
    """
    def __init__(self, store='', db=''):
        if store:
            self.store = store
        elif db:
            self.store = TracStore(db)
        self.series = []

    def getUsers(self, dateRange):
        all = [x.author for x in self.store.getRevisions(dateRange)]
        return list(set(all))

    def getRevisionStats(self, result):
        changedFiles = set(result.nodeChanges.values(NodeChange.path))
        branches = set(getBranches(changedFiles))
        ticketNumbers = parseTicketsFromBranches(branches)
        return {
            'author': result.author,
            'message': result.message,
            'changedFiles': changedFiles,
            'branches': branches,
            'ticketNumbers': ticketNumbers}

    def getTicketStats(self, result):
        closed = len([x for x in result.values(TicketChange.newvalue)
                     if x == 'closed'])
        return {
            'closed': closed,
            'updates': result.count()}

    def getAllStats(self, dateRange, userName=''):
        """
        Revision stats:
         * commits per time period in date range
         * user in each commit
         * branch names in each commit
         * associated ticket numbers in each commit
        """
        commitResults = self.store.getRevisions(dateRange, userName)
        ticketResults = self.store.getTickets(dateRange, userName)
        stats = {}
        commits = []
        for result in commitResults:
            commits.append(self.getRevisionStats(result))
        users = list(set([x['author'] for x in commits]))
        users.sort()
        branches = set()
        tickets = set()
        branchCreates = 0;
        fileCount = 0;
        for commit in commits:
            [branches.add(x) for x in commit['branches']]
            [tickets.add(x) for x in commit['ticketNumbers']]
            msg = commit['message']
            if msg and msg.startswith('Branching'):
                branchCreates += 1
            fileCount += len(commit['changedFiles'])
        branches = list(branches)
        branches.sort()
        tickets = list(tickets)
        def sortTickets(a,b):
            try:
                return cmp(int(a.strip('#')), int(b.strip('#')))
            except:
                return cmp(a.strip('#'), b.strip('#'))
        tickets.sort(sortTickets)
        stats['commits'] = commits
        stats['commitCount'] = len(commits)
        stats['users'] = users
        stats['userCount'] = len(users)
        stats['branches'] = branches
        stats['branchEdits'] = len(branches)
        stats['branchCreates'] = branchCreates
        stats['tickets'] = tickets
        stats['fileCount'] = fileCount
        stats['ticketCount'] = len(tickets)
        ticketStats = self.getTicketStats(ticketResults)
        stats['ticketUpdates'] = ticketStats['updates']
        stats['ticketCloses'] = ticketStats['closed']
        return stats

    @staticmethod
    def getStatTypes():
        """
        These are the stats and rates indices whose associated values are
        numbers and thus usable in computations.
        """
        return ['commitCount', 'userCount', 'branchEdits', 'branchCreates',
            'fileCount', 'ticketCount', 'ticketCloses', 'ticketUpdates']

    def getRatesChange(self, dateRange, thisStats):
        rates = {}
        thisRange = makeEpochDateRange(dateRange)
        lastRange = getLastTimePeriod(thisRange)
        lastStats = self.getAllStats(lastRange)
        ratableIndices = self.getStatTypes()
        for index in ratableIndices:
            changeInY = thisStats[index] - lastStats[index]
            # let's put the time in units of days instead of seconds, since
            # that's the smallest resolution we will be interested in
            changeInT = (thisRange[1] - lastRange[0])/(24 * 60 * 60)
            rates[index] = round(float(changeInY)/float(changeInT), 2)
        return rates

    def _makeHeader(self, text, char='='):
        header = char*len(text)+'\n'
        header += text+'\n'
        header += char*len(text)+'\n'
        return header

    def getStatSummary(self, dateRange, userName=''):
        stats = self.getAllStats(dateRange, userName)
        rates = self.getRatesChange(dateRange, stats)
        summary = ''
        if userName:
            summary += self._makeHeader(userName, '-')
        else:
            summary += self._makeHeader('Totals', '=')
        summary += "Number of commits: %i (%s)\n" % (
            stats['commitCount'], rates['commitCount'])
        if not userName:
            summary += "Committing users: %s (%s)\n" % (
                ', '.join(stats['users']), rates['userCount'])
        summary += "Total files changed: %i (%s)\n" % (
            stats['fileCount'], rates['fileCount'])
        summary += "Total branches created: %i (%s)\n" % (
            stats['branchCreates'], rates['branchCreates'])
        summary += "Total branches worked on: %i (%s)\n" % (
            stats['branchEdits'], rates['branchEdits'])
        summary += "Total tickets worked on: %i (%s)\n" % (
            stats['ticketCount'], rates['ticketCount'])
        summary += "Total ticket updates: %s (%s)\n" % (
            stats['ticketUpdates'], rates['ticketUpdates'])
        summary += "Total tickets closed: %s (%s)\n" % (
            stats['ticketCloses'], rates['ticketCloses'])
        summary += "All branches: %s\n" % ', '.join(stats['branches'])
        summary += "All ticket numbers: %s\n" % ', '.join(stats['tickets'])
        return summary

    def writeStatSummary(self, filename, dateRange, userName=''):
        pathname = os.path.basename(filename)
        if not os.path.exists(pathname):
            os.makedirs(pathname)
        summary = self.getStatSummary(dateRange, userName)
        if not userName:
            userName = 'the collective'
        print "Writing textual summary file for %s ..." % userName
        fh = open(filename, 'w+')
        fh.write(summary)
        fh.close()

    def getAllUsersStatSummary(self, dateRange):
        output = self._makeHeader('Per-user Breakdown', '=')
        output += '\n'
        for user in self.getUsers(dateRange):
            output += self.getStatSummary(dateRange, user)
            output += '\n'
        return output

    def printCommitStats(self, dateRange):
        print self.getStatSummary(dateRange)
        print self.getAllUsersStatSummary(dateRange)

    @staticmethod
    def getTimePoints(dateRange, step='day'):
        """
        step - the resoluton of the time increment; legal values are:
                'day', 'week'
        """
        if step not in ['day', 'week']:
            raise ValueError, "Invalid step '%s'." % step
        epochRange = makeEpochDateRange(dateRange)
        values = []
        step = eval(step)
        delta = epochRange[1] - epochRange[0]
        currentEpoch = epochRange[1] + second
        while currentEpoch > epochRange[0]:
            if not currentEpoch > epochRange[-1]:
                values.append(currentEpoch)
            currentEpoch -= step
        values.append(epochRange[0])
        return values

    def createStatsSeries(self, dateRange, step='day', userName=''):
        series = []
        timePoints = self.getTimePoints(dateRange, step)
        timePoints.reverse()
        for index, point in enumerate(timePoints):
            try:
                incrementRange = [point, timePoints[index+1]]
            except IndexError:
                # we hit the end of the time series
                pass
            allStats = self.getAllStats(incrementRange, userName)
            numericStats = []
            stats = {}
            for key, val in allStats.items():
                if key in self.getStatTypes():
                    stats[key] = val
            rates = self.getRatesChange(incrementRange, stats)
            data = {
                'time': point,
                'datetime': convertEpoch2DT(point),
                'stats': stats,
                'rates': rates}
            series.append(data)
        return series

    def setStatsSeries(self, dateRange, step='day', userName=''):
        self.series = self.createStatsSeries(dateRange, 'day', userName)
        
    def getSeriesData(self, type='', key=''):
        """
        A simple utillity method for extracting parts of the TracStats.series
        data structure.
        """
        if key:
            return [x[type][key] for x in self.series]
        return [x[type] for x in self.series]

class DayTicker(Locator):

    def __call__(self):
        start, end = self.dataInterval.get_bounds()
        middle = list(xrange(int(start + .5 * day), int(end + .5 * day), day))
        ticks = [start + .5 * day] + middle
        return ticks

class WeekTicker(Locator):
 
    def __call__(self):
        start, end = self.dataInterval.get_bounds()
        middle = list(xrange(int(start), int(end), week))
        ticks = [start] + middle
        return ticks

class MonthTicker(Locator):

    def __call__(self):
        start, end = self.dataInterval.get_bounds()
        middle = list(xrange(int(start), int(end), int(month)))
        ticks = [start] + middle
        return ticks

class YearTicker(Locator):

    def __call__(self):
        start, end = self.dataInterval.get_bounds()
        middle = list(xrange(int(start), int(end), int(year)))
        ticks = [start] + middle
        return ticks

class EpochDayOfWeekFormatter(Formatter):

    def __call__(self, x, pos=None):
        dt = convertEpoch2DT(x)
        return dt.strftime('%a')

class EpochWeekFormatter(Formatter):
                                
    def __call__(self, x, pos=None):
        dt = convertEpoch2DT(x)
        return dt.strftime('%a, %b %d')

class EpochMonthFormatter(Formatter):

    def __call__(self, x, pos=None):
        dt = convertEpoch2DT(x)
        return dt.strftime('%b')

class EpochYearFormatter(Formatter):

    def __call__(self, x, pos=None):
        dt = convertEpoch2DT(x)
        return dt.strftime('%Y')

class Subplot(axes.SubplotBase, axes.Axes):

    def __init__(self, fig, rect, *args, **kwargs):
        axes.SubplotBase.__init__(self, fig, 111)
        axes.Axes.__init__(self, fig, rect)

class GraphWrapper(object):

    def __init__(self, parent, size='large'):
        apsectRatio = 1.77
        dpi = 72
        def getSizeInches(width):
            w = width/dpi
            h = w/apsectRatio
            return (w, h)
        if size == 'medium':
            width = 300
            figsize = getSizeInches(width)
        elif size == 'small':
            width = 150
            figsize = getSizeInches(width)
        elif size == 'large':
            width = 800
            figsize = getSizeInches(width)
        elif size == 'thumb':
            width = 75
            figsize = getSizeInches(width)
        elif size == 'massive':
            width = 1024
            figsize = getSizeInches(width)
        self.statgraphs = parent
        self.figure = Figure(figsize=figsize, dpi=dpi)
        self.canvas = FigureCanvasAgg(self.figure)

        rect = [0.2, 0.3, 0.7, 0.6]
        subplot = Subplot(self.figure, rect)
        self.figure.axes.append(subplot)
        self.figure._axstack.push(subplot)
        self.ax = self.figure.sca(subplot)
        #self.ax = self.figure.add_subplot(111)
        #self.ax = self.figure.add_axes([0.2, 0.3, 0.7, 0.6], frame_on=False)
        self.majorLocator = None
        self.majorFormatter = None
        self.fileName = ''

    def label(self, title, x, y):
        self.ax.set_title(title)
        self.ax.set_xlabel(x)
        self.ax.set_ylabel(y)

    def configureAxes(self):
        timePoints = self.statgraphs.timePoints
        delta = self.statgraphs.delta
        self.ax.set_xlim(timePoints[0], timePoints[-1] + delta)
        res = self.statgraphs.res
        # get tick locator
        if res['ticks'] == 'day':
            self.majorLocator = DayTicker()
        elif res['ticks'] == 'week':
            self.majorLocator = WeekTicker()
        elif res['ticks'] == 'month':
            self.majorLocator = MonthTicker()
        elif res['ticks'] == 'year':
            self.majorLocator = YearTicker()
        # get date formatter
        if res['date'] == 'day':
            self.majorFormatter = EpochDayOfWeekFormatter()
        elif res['date'] == 'week':
            self.majorFormatter = EpochWeekFormatter()
        elif res['date'] == 'month':
            self.majorFormatter = EpochMonthFormatter()
        elif res['date'] == 'year':
            self.majorFormatter = EpochYearFormatter()
        self.ax.xaxis.set_major_locator(self.majorLocator)
        self.ax.xaxis.set_major_formatter(self.majorFormatter)
        if res['ticks'] == 'day':
            for label in self.ax.get_xticklabels():
                label.set_rotation(75)
                label.set_fontsize(8)
        self.ax.grid(False)
        self.ax.yaxis.grid(True)

    @staticmethod
    def getSizes():
        return ['full', 'medium', 'small', 'thumb']

    def write(self, fileName='', size='full'):
        assert size in self.getSizes()
        kwds = {}
        if size == 'full':
            dpi = None
        elif size == 'medium':
            dpi = 60
        elif size == 'small':
            dpi = 36
        elif size == 'thumb':
            dpi = 12
        if self.fileName and not fileName:
            filename = self.fileName
        if dpi:
            kwds['dpi'] = dpi
        path = os.path.dirname(fileName)
        try:
            os.makedirs(path)
        except OSError:
            # already exists
            pass
        self.canvas.figure.savefig(fileName, **kwds)

class TracStatGraphs(object):

    def __init__(self, store, outDir='/tmp'):
        self.stats = TracStats(store)
        self.dateRange = ''
        self.outDir = outDir
        self.sizes = []
        self.getSizes = lambda: None
        self.timePoints = []
        self.adjustedTimePoints = []
        self.delta = 0.0
        self.span = 0
        self.res = ''

    def setupTimePoints(self, timePoints):
        self.timePoints = timePoints
        self.adjustedTimePoints = timePoints
        self.delta = self.timePoints[1] - self.timePoints[0]
        self.span = self.timePoints[-1] - self.timePoints[0]
        self.res = self.getResolution()

    def getResolution(self):
        if not self.span > 45 * day:
            ticks = 'day'; date = 'week'; width = 'day'
        elif not self.span > 3 * month:
            ticks = 'week'; date = 'week'; width = 'day'
        elif not self.span > 1 * year + .5 * year:
            ticks = 'month'; date = 'month'; width = 'week'
        elif not self.span > 4 * year:
            ticks = 'year'; date = 'year'; width = 'week'
        else:
            ticks = 'year'; date = 'year'; width = 'month'
        return {'ticks': ticks, 'date': date, 'barWidth': width}

    def _adjustTimePoints(self, step):
        self.adjustedTimePoints = self.stats.getTimePoints(
            self.dateRange, step)
        self.adjustedTimePoints.reverse()
        tp = self.timePoints[-1]
        atp = self.adjustedTimePoints[-1]
        if atp < tp:
            delta = atp - self.adjustedTimePoints[-2]
            self.adjustedTimePoints.append(atp + delta)
        return self.adjustedTimePoints

    def _adjustData(self, data):
        zipped = zip(self.timePoints, data)
        newData = [0] * len(self.adjustedTimePoints)
        for index, newTime in enumerate(self.adjustedTimePoints):
            time = newTime
            try:
                nextNewTime = self.adjustedTimePoints[index + 1]
            except IndexError:
                # end of the line
                pass
            while time < nextNewTime:
                try:
                    time, yval = zipped.pop(0)
                except IndexError:
                    # popped the last value already
                    break
                if time < nextNewTime:
                    #newData[index] = newData[index] + yval
                    newData[index] += yval
                else:
                    newData[index + 1] += yval
        # make sure we got it right
        assert(sum(data) == sum(newData))
        return newData

    def correlateData(self, data):
        # check to see if time points need to be converted from days
        timePoints = self.timePoints
        step = self.res['barWidth']
        if eval(step) != day:
            if self.adjustedTimePoints == self.timePoints:
                # adjust timepoints
                timePoints = self._adjustTimePoints(step)
            else:
                # use already-calculated data
                timePoints = self.adjustedTimePoints
            # adjust data
            data = self._adjustData(data)
        return [timePoints, data]

    def createGraph(self, data, ylabel, xlabel='time', title='',
        subTitle='', graphType='bar', size='full'):
        if not title:
            title = '%s vs. %s' % (ylabel, xlabel)
        if subTitle:
            title = '%s\n%s' % (title, subTitle)
        graph = GraphWrapper(self)
        self.getSizes = graph.getSizes
        if graphType == 'bar':
            barWidth = eval(self.res['barWidth'])
            data.append(barWidth)
            try:
                graph.ax.bar(*data)
            except Exception, err:
                import pdb;pdb.set_trace()
        elif graphType == 'line':
            graph.ax.plot(linewidth=2, *data)
        graph.label(title, xlabel, ylabel)
        graph.configureAxes()
        fileName = re.sub('[. ]', '', title.split('\n')[0])
        graph.write("%s/%s_%s.png" % (self.outDir, fileName, size), size=size)

    def createCompositeGraph(self, yValues, ylabel, xlabel='time',
        title='', subTitle='', graphType='line', size='full'):
        if not title:
            title = 'Activity over Time'
        if subTitle:
            title = '%s\n%s' %  (title, subTitle)
        graph = GraphWrapper(self)
        self.getSizes = graph.getSizes
        graph.label(title, xlabel, ylabel)
        lines = []
        for dataSet in yValues:
            graphData = self.correlateData(dataSet)
            line = graph.ax.plot(linewidth=2, *graphData)
            lines.append(line)
        labels = [titleFromCamelCase(x) for x in
            self.stats.getStatTypes()]
        graph.ax.legend(lines, labels, 'best', shadow=True)
        graph.configureAxes()
        fileName = re.sub('[. ]', '', title.split('\n')[0])
        graph.write("%s/%s_%s.png" % (self.outDir, fileName, size), size=size)

    def createAllGraphs(self, dateRange, userName='', size='full'):
        self.dateRange = dateRange
        humanDR = ' to '.join([x.strip() for x in
            dateRange.split(',')]).replace('-', '.')
        self.stats.setStatsSeries(dateRange, 'day', userName)
        timePoints = self.stats.getSeriesData('time')
        self.setupTimePoints(timePoints)
        xlabel = 'Time'
        userTitle = ''
        if userName:
            userTitle = ' (%s)' % userName
        # composite stats graph
        title = 'Composite Stats%s' % userTitle
        statYVals =  [self.stats.getSeriesData('stats', key) for key in
            self.stats.getStatTypes()]
        compositeStatGraph = self.createCompositeGraph(statYVals,
            'Counts', xlabel, title=title, subTitle=humanDR, size=size)
        # composite rates graph
        title = 'Composite Rates of Change%s' % userTitle
        rateYVals =  [self.stats.getSeriesData('rates', key) for key in
            self.stats.getStatTypes()]
        compositeStatGraph = self.createCompositeGraph(rateYVals,
            'Changes', xlabel, title=title, subTitle=humanDR, size=size)
        if not userName:
            print "Creating collective graphs during %s (%s size) ..." % (
                dateRange, size)
        else:
            print "Creating graphs for %s during %s (%s size) ..." % (
                userName, dateRange, size)
        for key in self.stats.getStatTypes():
            print " * Creating %s graphs ..." % key
            statPlotPoints = self.correlateData(
                self.stats.getSeriesData('stats', key))
            ratePlotPoints = self.correlateData(
                self.stats.getSeriesData('rates', key))
            ylabel = titleFromCamelCase(key)
            userTitle = ''
            if userName:
                userTitle = ' (%s)' % userName
            title = "%s vs. %s%s" % (ylabel, xlabel, userTitle)
            self.createGraph(statPlotPoints, ylabel, title=title,
                             subTitle=humanDR, size=size)
            title = "%s Rate of Change%s" % (ylabel, userTitle)
            ylabel = 'Change in %s' % ylabel
            self.createGraph(ratePlotPoints, ylabel, title=title,
                             subTitle=humanDR, graphType='line', size=size)

    def createAllUserGraphs(self, dateRange, userName, size='full'):
        for user in self.stats.getUsers(dateRange):
            print "Creating graphs for %s ..." % user
            self.createAllGraphs(dateRange, user, size=size)

class TracStatViews(object):

    def __init__(self, store, outDir='/tmp'):
        self.outDir = outDir
        self.graphs = TracStatGraphs(store, outDir)
        self.stats = self.graphs.stats
        self.getUsers = self.stats.getUsers

    def getHTML(self, dateRange, userName=''):
        pass

    def writeHTML(self, filename, dateRange, userName=''):
        pass

    def createHTMLWeek(self, date, userName=''):
        """
        The week view returns a tabular representation of the week, with the
        following:
            * a list of days in a line
            * for every day that has data, the day block is clickable
            * below the tabular days, graphs for the collective are displayed 
              (stats and rates)
            * below the collective graphs, graphs per user are displayed
              (stats and rates)
            * below the graphs section, textual sumaries are displayed
        """

    def createHTMLMonth(self, date, userName=''):
        pass

    def createHTMLYear(self, date, userName=''):
        pass

    def createHTMLYearList(self, yearRange, userName=''):
        pass

    def createHTMLStatGraphs(self, dateRange, userName=''):
        self.createStatGraphs(dateRange, userName)
        if not userName:
            print "Creating collective HTML views ..."
        if not userName:
            for user in self.stats.getUsers(dateRange):
                print "Creating HTML views for %s ..." % user


